use simcore_rs::{Time, SimContext, Facility, RandomVar, Process};
use rand::{rngs::SmallRng, SeedableRng, Rng};
use std::cell::RefCell;

/// Globally shared data.
struct Shared {
	rng_a: RefCell<SmallRng>,
	rng_s: RefCell<SmallRng>,
	joe: Facility,
	wait_time: RandomVar
}

// helper constants
const SEED_A: u64  = 100000;
const SEED_S: u64  = 200000;

/// Customer process with access to the barber and a random processing delay.
struct Customer;

impl Customer {
	pub async fn actions(self, sim: SimContext<'_,Shared>) {
		// access the barber and record the time for the report
		let arrival_time = sim.now();
		sim.shared().joe.seize().await;
		sim.shared().wait_time.tabulate(sim.now() - arrival_time);
		
		// spend time
		sim.advance(
			sim.shared().rng_s.borrow_mut().gen_range(12.0..18.0)
		).await;
		// release the barber
		sim.shared().joe.release();
	}
}

async fn barbershop(sim: SimContext<'_,Shared>, duration: Time) {
	// activate a process to generate the customers
	sim.activate(async move {
		loop {
			sim.advance(
				sim.shared().rng_a.borrow_mut().gen_range(12.0..24.0)
			).await;
			if sim.now() >= duration { return; }
			sim.activate(Customer.actions(sim));
		}
	});
	
	// wait until the store closes
	sim.advance(duration).await;
	
	// finish processing the queue (no more customers arrive)
	sim.shared().joe.seize().await;
}

#[cfg(not(test))]
fn main() {
	let result = simcore_rs::simulation(
		// global data
		Shared {
			rng_a: RefCell::new(SmallRng::seed_from_u64(SEED_A)),
			rng_s: RefCell::new(SmallRng::seed_from_u64(SEED_S)),
			joe: Facility::new(),
			wait_time: RandomVar::new()
		},
		// simulation entry point
		|sim|
			Process::new(sim, barbershop(sim, 60.0*24.0*7.0*3.0))
	);
	
	// print statistics
	println!("wait_time: {:#.3}", result.wait_time);
}

#[cfg(test)]
criterion::criterion_main!(bench::benches);

#[cfg(all(test, windows))]
mod slx;

#[cfg(test)]
mod bench {
	use super::*;
	use criterion::{Criterion, BenchmarkId, BatchSize, PlotConfiguration,
	                AxisScale, SamplingMode, criterion_group};
	use std::time::Duration;
	
	mod odemx {
		use std::os::raw::c_double;
		
		#[link(name = "odemx", kind = "static")]
		extern {
			pub fn barbershop(duration: c_double);
		}
	}
	#[cfg(windows)]
	const SLX_PATH: &'static str = "C:/Wolverine/SLX/se64.exe";
	const RANGE: u32 =   10;
	const STEP: Time = 1000.0;
	
	fn barbershop_bench(c: &mut Criterion) {
		let mut group = c.benchmark_group("Barbershop");
		
		// set-up the benchmark parameters
		group.confidence_level(0.99);
		group.plot_config(
			PlotConfiguration::default()
				.summary_scale(AxisScale::Logarithmic)
		);
		group.sampling_mode(SamplingMode::Linear);
		//group.measurement_time(Duration::from_secs(300));
		
		if !cfg!(windows) {
			println!("Not running under Windows, skipping SLX benchmarks!");
		}
		
		// vary in the length of the simulation run
		for sim_duration in (0..RANGE).map(|c| Time::from(1 << c)*STEP) {
			#[cfg(windows)] {
				let duration = sim_duration.to_string();
				let args = [
					"/silent",
					"/stdout",
					"/noicon",
					"/nowarn",
					"/noxwarn",
					"/#BENCH",
					"slx\\barbershop.slx",
					duration.as_str()
				];
				
				// benchmark the SLX implementation
				group.bench_function(
					BenchmarkId::new("SLX", sim_duration),
					|b| b.iter_custom(|iters|
						slx::slx_bench(
							SLX_PATH,
							&args,
							iters as usize
						).expect("couldn't benchmark the SLX program")
					)
				);
			}
			
			// benchmark the C++ implementation
			group.bench_function(
				BenchmarkId::new("ODEMx", sim_duration),
				|b| b.iter(
					|| unsafe { odemx::barbershop(sim_duration as _) }
				)
			);
			
			// benchmark the Rust implementation
			group.bench_function(
				BenchmarkId::new("Rust", sim_duration),
				|b| b.iter_batched(
					|| Shared {
						rng_a: RefCell::new(SmallRng::seed_from_u64(SEED_A)),
						rng_s: RefCell::new(SmallRng::seed_from_u64(SEED_S)),
						joe: Facility::new(),
						wait_time: RandomVar::new()
					},
					|shared| simcore_rs::simulation(
						shared,
						|sim| sim.process(barbershop(sim, sim_duration))
					),
					BatchSize::SmallInput
				)
			);
		}
		
		group.finish();
	}
	
	criterion_group!(benches, barbershop_bench);
}
